{%MainUnit ../suggestedit.pas

*****************************************************************************
                               suggestedit.inc
                               ---------------
                            SuggestEdit Component

                       Initial Revision  : 08.12.2015

*****************************************************************************
 This file is part of the SuggestEdit Component.
 Copyright (c) 2015 by Michael W. Vogel

 See the file COPYING.modifiedLGPL.txt, included in this distribution,
 for details about the license.
*****************************************************************************
}

procedure TCustomSuggestEdit.FBoxPreviewFirstItem(
  const S: string);
begin
  Text := S;
  CaretPos := Point(Length(FInsertedText), 0);
end;

procedure TCustomSuggestEdit.FBoxSelectionChange(const aValue: string);
var
  s: string;
begin
  if not Assigned(FBox) then Exit;
  FKeyForListBox := True;
  if FBox.ItemSelected then s := aValue
                       else s := FInsertedText;
  Text := s;
  SelStart := Length(s);
end;

procedure TCustomSuggestEdit.FBoxSelectItem(const s: String);
begin
  Text := s;
  FInsertedText := s;
  if Assigned(FBox) then
    FBox.Visible := False;
end;

function TCustomSuggestEdit.GetSortingOrder: TSortingOrder;
begin
  Result := TSuggestStrings(Items).SortingOrder;
end;

procedure TCustomSuggestEdit.SetSortingOrder(AValue: TSortingOrder);
begin
  TSuggestStrings(Items).SortingOrder := AValue;
end;

procedure TCustomSuggestEdit.Change;
var
  aLenght: Integer;
begin
  inherited Change;
  if FKeyForListBox then Exit;
  if not FKeyIsDown then Exit;
  if ssUpdatePreviewFirstItem in State then Exit;
  aLenght := CaretPos.x;
  FInsertedText := Copy(Text, 1, aLenght);
  if Assigned(FBox) then FBox.FillBox(FInsertedText);
end;

procedure TCustomSuggestEdit.DeleteItem;
var
  Deleted: Boolean;
begin
  if Assigned(FBox) then begin
    FKeyForListBox := True;
    FKeyIsDown := False;
    Deleted := True;
    if Assigned(FOnDeleteItem) then
      FOnDeleteItem(Self, FBox.SelectedItem, Deleted);
    if Deleted then
    begin
      Items.Delete(Items.IndexOf(FBox.SelectedItem));
      FBox.DeleteIndex;
    end;
  end;
end;

procedure TCustomSuggestEdit.DoEnter;
begin
  if Assigned(FOnFillItems) then
    FOnFillItems(Self);
  inherited DoEnter;
end;

procedure TCustomSuggestEdit.DoExit;
begin
  inherited DoExit;
  if Assigned(FBox) then
    FBox.Visible := False;
end;

procedure TCustomSuggestEdit.DoFocus;
begin
  SetFocus;
  CaretPos := Point(Length(FInsertedText), 0);
end;

procedure TCustomSuggestEdit.KeyDown(var Key: Word;
  Shift: TShiftState);
begin
  FKeyIsDown := True;
  case Key of
    VK_UP:
      begin
        FKeyForListBox := True;
        if Assigned(FBox) and FBox.Showing then
          FBox.MoveItemIndexUp;
        Key := 0;
      end;
    VK_DOWN:
      begin
        FKeyForListBox := True;
        if Assigned(FBox) then
          if FBox.Showing then FBox.MoveItemIndexDown
                          else FBox.FillBox(Text);
        Key := 0;
      end;
    VK_PRIOR: // Page Up
      begin
        FKeyForListBox := True;
        if Assigned(FBox) and FBox.Showing then
          FBox.MoveItemIndexPageUp;
        Key := 0;
      end;
    VK_NEXT: // Page Down
      begin
        FKeyForListBox := True;
        if Assigned(FBox) then
          if FBox.Showing then FBox.MoveItemIndexPageDown
                          else FBox.FillBox(Text);
        Key := 0;
      end;
    VK_RETURN:
      if Assigned(FBox) then begin
        if FBox.ItemSelected and FKeyForListBox then
          FBoxSelectItem(FBox.SelectedItem)
        else
          FBox.Visible := False;
      end;
    VK_TAB:
      if Assigned(FBox) then begin
        if FBox.Showing and FBox.ItemSelected then
        begin
          Text := FBox.SelectedItem;
          FInsertedText := FBox.SelectedItem;
        end;
        FBox.Visible := False;
      end;
    VK_ESCAPE:
      begin
        if Assigned(FBox) and FBox.Showing then
        begin
          FKeyForListBox := True;
          FBox.Visible := False;
          Text := FInsertedText;
          SelStart := Length(FInsertedText);
        end;
        Key := 0;
      end;
    VK_DELETE:
      if (Shift = [ssCtrl]) and Assigned(FBox) and FBox.Visible and FBox.ItemSelected then
        DeleteItem;
    else
      FKeyForListBox := False;
  end;
  inherited KeyDown(Key, Shift);
end;

procedure TCustomSuggestEdit.KeyUp(var Key: Word;
  Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift);
  FKeyIsDown := False;
end;

procedure TCustomSuggestEdit.SetColor(Value: TColor);
begin
  inherited SetColor(Value);
  if Assigned(FBox) then
    FBox.Color := Value;
end;

procedure TCustomSuggestEdit.SetParent(NewParent: TWinControl);
begin
  inherited SetParent(NewParent);
  if Assigned(FBox) then
    FBox.Parent := Nil;
end;

constructor TCustomSuggestEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBox := Nil;
  FItemsUpDating := 0;
  FKeyForListBox := False;
  FKeyIsDown := False;
  State := [];
  MaxBoxHeight := 200;
  MaxBoxPreviewCount := -1;
  if csDesigning in ComponentState then Exit;
  FBox := TSuggestEditListboxForm.Create(Self);
  FBox.OnSelectionChange := FBoxSelectionChange;
  FBox.OnSelectItem := FBoxSelectItem;
  FBox.OnPreviewFirstItem := FBoxPreviewFirstItem;
end;

procedure TCustomSuggestEdit.EditingDone;
var
  Inserting: Boolean;
begin
  FKeyIsDown := False;
  if (not Assigned(FBox)) or (not FBox.Visible) then
  begin
    inherited EditingDone;
    Inserting := (Text <> '') and (Items.IndexOf(Text) < 0);
    if Inserting and Assigned(FOnNewItem) then
      FOnNewItem(Self, Text, Inserting);
    if Inserting then
      Items.Add(Text);

    if SortingOrder = soLastUsed then
      TSuggestStrings(Items).SetLastUsedItem(Text);
  end;
end;

function TCustomSuggestEdit.Focused: Boolean;
begin
  Result := inherited Focused;
  if Assigned(FBox) then Result := Result or FBox.Focused;
end;

procedure TCustomSuggestEdit.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, Height);
  if Assigned(FBox) then
    FBox.ResizeForm;
end;



