{
*****************************************************************************
                               suggestedit.pas
                               ---------------
                            SuggestEdit Component

                       Initial Revision  : 08.12.2015

*****************************************************************************
 This file is part of the SuggestEdit Component.
 Copyright (c) 2015 by Michael W. Vogel

 See the file COPYING.modifiedLGPL.txt, included in this distribution,
 for details about the license.
*****************************************************************************

 How to use
 ----------

 You can insert some data from database or local file after creating your
 SuggestEdit. Your SuggestEdit holds its entries all the time, so you don't need
 a extra container for that data. Before destroying your SuggestEdit, you can
 save the entries to a local file.

 If you work with a little database you can also do:

 - OnEnter:       Fill data to be selected from.
                  E.G. ASuggestEdit.Items.Assign(SomeDataStringList);

 - OnNewItem:     Insert the new entry in your local data container / database
                  E.G. SomeDataStringList.Add(AText);

 - OnDeleteItem:  Delete the entry in your local data container / database and
                  fill your TSuggestEdit with the local data container / database
                  E.G. SomeDataStringList.Delete(SomeDataStringList.IndexOf(AText));
                       ASuggestEdit.Items.Assign(SomeDataStringList);
}

unit SuggestEdit;

{$mode Delphi}{$H+}

interface

uses
  Classes, sysutils, math, fgl, StdCtrls, Controls, LCLType, LCLProc,
  Graphics, LCLIntf, Forms, LazUTF8, LResources, ExtCtrls;

type

  TCompareOption = (coCaseSensitive, coAnywhere);
  TCompareOptions = Set of TCompareOption;

  TSuggesteditState = (ssUpdatePreviewFirstItem);
  TSuggesteditStates = Set of TSuggesteditState;

  TStringValidEvent = procedure(Sender: TObject; const AText: String; var Valid: Boolean) of object;

  { TSuggestStringsItem }

  TSuggestStringsItem = record
    Str: String;
    Obj: TObject;
    LastUsed: DWord;
    class operator Equal(Item1, Item2: TSuggestStringsItem): Boolean;
  end;

  TSuggestStringsItems = class( specialize TFPGList<TSuggestStringsItem> );

  TSuggestStrings = class;
  TSuggestStringsSortCompare = function(List: TSuggestStrings; Index1, Index2: Integer): Integer;
  TSortingOrder = (soNone, soAsc, soDesc, soLastUsed);

  { TSuggestStrings }

  TSuggestStrings = class(TStrings)
  private
    FList: TSuggestStringsItems;
    FSortingFunction: TSuggestStringsSortCompare;
    FSortingOrder: TSortingOrder;
    function GetLastUsed(Index: Integer): DWord;
    procedure SetLastUsed(Index: Integer; AValue: DWord);
    procedure SetSortingOrder(AValue: TSortingOrder);
  protected
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
  protected
    property SortingFunction: TSuggestStringsSortCompare read FSortingFunction write FSortingFunction;
    property SortingOrder: TSortingOrder read FSortingOrder write SetSortingOrder default soNone;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure SetLastUsedItem(ItemText: String);
    procedure Sort;
  public
    property LastUseds[Index: Integer]: DWord read GetLastUsed write SetLastUsed;
  end;

  { TVirtualSuggestEdit }

  TVirtualSuggestEdit = class(TCustomEdit)
  private
    FCompareOptions: TCompareOptions;
    FItems: TStrings;                                          // TSuggestStrings
    FMaxBoxHeight: Integer;
    FMaxBoxPreviewCount: Integer;
    FSelectFirstItem: Boolean;
    FState: TSuggesteditStates;
    procedure SetCompareOptions(AValue: TCompareOptions);
    procedure SetSelectFirstItem(AValue: Boolean);
  protected
    procedure DoFocus; virtual; abstract;
  protected
    property CompareOptions: TCompareOptions read FCompareOptions write SetCompareOptions;
    property Items: TStrings read FItems;
    property MaxBoxHeight: Integer read FMaxBoxHeight write FMaxBoxHeight default 200;
    property MaxBoxPreviewCount: Integer read FMaxBoxPreviewCount write FMaxBoxPreviewCount default -1;
    property SelectFirstItem: Boolean read FSelectFirstItem write SetSelectFirstItem default false;
    property State: TSuggesteditStates read FState write FState;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  { TSuggestEditListboxForm }

  TSuggestEditListboxForm = class(TForm)
  private
    FBorderHeight: Integer;
    FCurrentEditText: String;
    FFirstShow: Boolean;
    FItemHeight: Integer;
    FItemsDummy: TStringList;
    FItemShowingCount: Integer;
    FListBox: TListBox;
    FOnPreviewFirstItem: TGetStrProc;
    FOnSelectionChange: TGetStrProc;
    FOnSelectItem: TGetStrProc;
    FParentOKTimer: TTimer;
    FParentPoint: TPoint;
    FSuggestEdit: TVirtualSuggestEdit;
    procedure ClickItem({%H-}Sender: TObject);
    procedure FillItemsDummy;
    procedure FocusEdit({%H-}Sender: TObject);
    function GetItemSelected: Boolean;
    function GetSelectedItem: String;
    procedure InitializeParentOK;
    procedure MouseMoveItems({%H-}Sender: TObject; {%H-}Shift: TShiftState; X, Y: Integer);
    procedure CheckParentOK({%H-}Sender: TObject);
    procedure PrepareFirstShow;
    procedure SelectItem;
  protected
    procedure SetColor(Value: TColor); override;
    procedure SetVisible(Value: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    constructor CreateNew(AOwner: TComponent; Num: Integer = 0); override;
    destructor Destroy; override;
    procedure DeleteIndex;
    procedure FillBox(const s: String);
    procedure FocusOwner({%H-}Data: PtrInt);
    function GetSelectedText: String;
    procedure MoveItemIndexDown;
    procedure MoveItemIndexPageDown;
    procedure MoveItemIndexPageUp;
    procedure MoveItemIndexUp;
    procedure SelectionChanged({%H-}Sender: TObject; {%H-}User: boolean);
    procedure ResizeForm;
  public
    property ItemSelected: Boolean read GetItemSelected;
    property ListBox: TListBox read FListBox;
    property OnPreviewFirstItem: TGetStrProc read FOnPreviewFirstItem write FOnPreviewFirstItem;
    property OnSelectionChange: TGetStrProc read FOnSelectionChange write FOnSelectionChange;
    property OnSelectItem: TGetStrProc read FOnSelectItem write FOnSelectItem;
    property SelectedItem: String read GetSelectedItem;
  end;

  { TCustomSuggestEdit }

  TCustomSuggestEdit = class(TVirtualSuggestEdit)
  private
    FBox: TSuggestEditListboxForm;
    FKeyForListBox: Boolean;
    FKeyIsDown: Boolean;
    FInsertedText: TCaption;
    FItemsUpDating: Integer;
    FOnDeleteItem: TStringValidEvent;
    FOnFillItems: TNotifyEvent;
    FOnNewItem: TStringValidEvent;
    procedure FBoxPreviewFirstItem(const S: string);
    procedure FBoxSelectionChange(const aValue: string);
    procedure FBoxSelectItem(const s: String);
    function GetSortingOrder: TSortingOrder;
    procedure SetSortingOrder(AValue: TSortingOrder);
  protected
    procedure Change; override;
    procedure DeleteItem;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure DoFocus; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure SetColor(Value: TColor); override;
    procedure SetParent(NewParent: TWinControl); override;
  protected
    property OnDeleteItem: TStringValidEvent read FOnDeleteItem write FOnDeleteItem;
    property OnFillItems: TNotifyEvent read FOnFillItems write FOnFillItems;
    property OnNewItem: TStringValidEvent read FOnNewItem write FOnNewItem;
    property SortingOrder: TSortingOrder read GetSortingOrder write SetSortingOrder default soNone;
  public
    constructor Create(AOwner: TComponent); override;
    procedure EditingDone; override;
    function Focused: Boolean; override;
    procedure SetBounds(ALeft, ATop, AWidth, {%H-}AHeight: Integer); override;
  public
    property AutoSelected default False;
    property AutoSize default True;
    property BorderStyle default bsSingle;
    property CaretPos: TPoint read GetCaretPos write SetCaretPos;
    property NumbersOnly: Boolean read GetNumbersOnly write SetNumbersOnly default false;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default false;
    property SelLength: integer read GetSelLength write SetSelLength;
    property SelStart: integer read GetSelStart write SetSelStart;
    property SelText: String read GetSelText write SetSelText;
    property TabStop default true;
  end;

  TSuggestEdit = class(TCustomSuggestEdit)
  published
    property Action;
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property AutoSelect;
    property BidiMode;
    property BorderSpacing;
    property BorderStyle;
    property CharCase;
    property Color;
    property CompareOptions;
    property Constraints;
    property Cursor;
    property DragCursor;
    property DragKind;
    property DragMode;
    property EchoMode;
    property Enabled;
    property Font;
    property HideSelection;
    property Items;
    property MaxBoxHeight;
    property MaxBoxPreviewCount;
    property MaxLength;
    property NumbersOnly;
    property ParentBidiMode;
    property OnChange;
    property OnChangeBounds;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDeleteItem;
    property OnDragDrop;
    property OnDragOver;
    property OnEditingDone;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnFillItems;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnNewItem;
    property OnResize;
    property OnStartDrag;
    property OnUTF8KeyPress;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property SelectFirstItem;
    property SortingOrder;
    property ShowHint;
    property TabStop;
    property TabOrder;
    property Text;
    property TextHint;
    property TextHintFontColor;
    property TextHintFontStyle;
    property Visible;
  end;

procedure register;

implementation

{$I suggeststrings.inc}
{$I virtualsuggestedit.inc}
{$I suggesteditlistboxform.inc}
{$I suggestedit.inc}

procedure register;
begin
  RegisterComponents('Misc', [TSuggestEdit]);
end;

initialization

{$I suggestedit_package.lrs}

end.

